#!/usr/bin/perl -w
#######################################################
# scanAlert.pl
# Written by: Brandon Zehm <caspian@dotconf.net>
#
# License:
#
# scanAlert.pl (hereafter referred to as "program") is free software;
#   you can redistribute it and/or modify it under the terms of the GNU General
#   Public License as published by the Free Software Foundation; either version
#   2 of the License, or (at your option) any later version.
# Note that when redistributing modified versions of this source code, you
#   must ensure that this disclaimer and the above coder's names are included
#   VERBATIM in the modified code.
#
# Disclaimer:
# This program is provided with no warranty of any kind, either expressed or
#   implied.  It is the responsibility of the user (you) to fully research and
#   comprehend the usage of this program.  As with any tool, it can be misused,
#   either intentionally (you're a vandal) or unintentionally (you're a moron).
#   THE AUTHOR(S) IS(ARE) NOT RESPONSIBLE FOR ANYTHING YOU DO WITH THIS PROGRAM
#   or anything that happens because of your use (or misuse) of this program,
#   including but not limited to anything you, your lawyers, or anyone else
#   can dream up.  And now, a relevant quote directly from the GPL:
#
#                           NO WARRANTY
#
#  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
# FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
# OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
# PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
# OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
# TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
# PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
# REPAIR OR CORRECTION.
#
#   ---
#
# Whee, that was fun, wasn't it?  Now let's all get together and think happy
#   thoughts - and remember, the same class of people who made all the above
#   legal spaghetti necessary are the same ones who are stripping your rights
#   away with DVD CCA, DMCA, stupid software patents, and retarded legal
#   challenges to everything we've come to hold dear about our internet.
#   So enjoy your dwindling "freedom" while you can, because 1984 is coming
#   sooner than you think.  :[
#
#######################################################
#
# Purpose: To create a program that is capable of monitoring iptables 
#          generated syslog messages, and detecting port scans.
#          
#
# Layout:  scanAlert is designed to recieve messages from syslogd via a fifo.
#          Syslogd needs to be configured to log iptables messages to that fifo.
#          scanAlert will monitor all the iptables messages passed in via that
#          fifo and generate an alert message (and send it to syslogd) when 
#          a portscan is detected.  
#          
# Features:
#       - Monitor kernel ipfilter generated messages and watch for port scans
#       - HUP signal is intercepted correctly and causes scanAlert to reload
#         its configuration and refresh all filehandles.
#       - Writes its state data to a file so that if it is killed and restarted
#         it knows what the status of things were when it shut down.
#       - Debugging mode with a few different debugging levels (enabled via a 
#         variable at the top of this script)
#          
# Changelog:         
#   08/22/2002 - v1.00-RC1
#       - Fixed a split to @_
#       - Fixed problem with scanAlert dieing after receiving a HUP signal.
#       - Fixed daemon mode so it forks to the background unless debugging is enabled.
#       - Enabled the log file functionality and made it log everything there by default.
#          
#   08/31/2002 - v1.00-RC2
#       - If not command line arguments are given help is printed
#       - Misc cleanup
#
use strict;
use IO::Socket;

## Global Variable(s)
my %conf = (
    "name"                 => $0,                                ## The name of this program
    "version"              => 'v1.00-RC2',                       ## The version of this program
    "authorName"           => 'Brandon Zehm',                    ## Author's Name
    "authorEmail"          => 'caspian@dotconf.net',             ## Author's Email Address
    "debug"                => 0,                                 ## Default debug level
    "configurationFile"    => '/etc/scanAlert.conf',             ## Default configuration file
    
    ## The following entries are used later in the program
    "mode"                 => "",                                ## either "", "daemon", or "running"
    "logging"              => "",                                ## If this is true the printmsg function prints to the log file
    
    ## The folling entries get their values from the configuration file
    "stateDate"            => "",
    "fifo"                 => "",
    "portScanThreshold"    => "",
    "hostExpiration"       => "",
    
);
$conf{"name"} =~ s/(.)*[\/,\\]//;                                ## Remove path from filename
$0 = "[$conf{'name'}]";

## Database of connection attempts
my %connectionHash = ();

## Read from the configuration file
my @ignorePortsTCP = ();
my @ignorePortsUDP = (); 
my @ignoreDstHosts = ();
my @ignoreSrcHosts = ();

## Set a default value for $conf{'debug'}
$conf{'debug'} = 0 unless ($conf{'debug'});









#############################
##
##      MAIN PROGRAM
##
#############################

## Initialize
initialize();

## Process Command Line
processCommandLine();

## Read Configuration File
readConfigurationFile($conf{'configurationFile'});
  
###################################################
##  Verify required variables are set correctly  ##
###################################################
if (1) {
  my @required = (
                    'stateData',
                    'fifo',
                    'portScanThreshold',
                    'hostExpiration',
  );
  foreach (@required) {
      if (!$conf{$_}) {
          quit("ERROR: Value [$_] was not defined in the configuration file!", 1);
      }
  }
};

## Open log file
if (openLogFile($conf{'logFile'})) { quit("ERROR: Opening the file [$conf{'logFile'}] returned the error: $!", 1); }

## Open fifo
if (openFIFO($conf{'fifo'})) { quit("ERROR: Connecting to fifo [$conf{'fifo'}] returned the error: $!", 1); }



## Read the stateData file into %connectionHash
if ( -f $conf{'stateData'} ) {
    if (open (FILE, $conf{'stateData'})) {
        my $key = "";
        my $value = "";
        while (<FILE>) {
            chomp $_;
            ($key, $value) = split(/\s+=>\s+/, $_);
            $connectionHash{$key} = $value if ($value);
        }
        close FILE;
    }
}
else {
    printmsg("NOTICE: The file [$conf{'stateData'}] does not exist.  Creating it now with mode [0600].", 0);
    open (FILE, ">$conf{'stateData'}");
    close FILE;
    chmod (0600, $conf{'stateData'});
}






#########################
##                     ##
##  MAIN PROGRAM LOOP  ##
##                     ##
#########################


if ($conf{"mode"} =~ /daemon/i) {
  
  ## Become deamon
  if ( ($conf{'debug'} >= 1) or (fork() == 0) ) {
      
      ## Close normal output methods since we won't be needing them in daemon mode
      close STDOUT unless ($conf{"debug"} >= 1);
      close STDERR unless ($conf{"debug"} >= 1);
      
      ## Set the mode to running so that messages get logged rather than printed to the screen
      $conf{"mode"} = "running";
      
      ## Setup HUP handler function - this will re-read the configuration file, re-open the fifo, log file, etc
      $SIG{'HUP'}   = sub { handleHUP(); };
      
      ## Set clean-up code to run every 5 minutes
      $SIG{'ALRM'}   = sub { handleALRM(); alarm 60; };
      alarm 60;
      
      ## Tell the user that we have started successfully
      printmsg("$$ - $conf{'name'} - daemon mode started successfully", 0);
      
      ## Forever
      while (1) {
          
          ## Read a line from the fifo
          my $message = <FIFO>;
          chomp $message;
          
          ## If this is an IPTables log entry
          if ($message =~ /SPT=/) {
              
              my $src = "";          ## Source host
              my $dst = "";          ## Destination host
              my $dpt = "";          ## Destination port
              my $protocol = "";     ## Protocol
              my $packetType = "";   ## Packet type
              my $ignorePacket = 0;  ## Ignore packet flag
              
              
              ## Get information about the packet
              ($src, $dst, $protocol, $dpt, $packetType) = getPacketInfo($message);
              
              
              ## Ignore packets tagged as "Expired Connection"
              unless ($ignorePacket) { if ($packetType eq "Expired Connection") { $ignorePacket = 1; } }
              
              ## Ignore TCP packets if the destination port is listed in the "ignorePortsTCP" section of the configuration file.
              unless ($ignorePacket) { if ($protocol eq "TCP") { foreach my $tmp (@ignorePortsTCP) { if ($tmp =~ /\-/) { (my $low, my $high) = split("-", $tmp); if (($dpt >= $low) and ($dpt <= $high)) { $ignorePacket = 1; } } elsif ($dpt == $tmp) { $ignorePacket = 1; } } } }
              
              ## Ignore UDP packets if the destination port is listed in the "ignorePortsUDP" section of the configuration file.
              unless ($ignorePacket) { if ($protocol eq "UDP") { foreach my $tmp (@ignorePortsUDP) { if ($tmp =~ /\-/) { (my $low, my $high) = split("-", $tmp); if (($dpt >= $low) and ($dpt <= $high)) { $ignorePacket = 1; } } elsif ($dpt == $tmp) { $ignorePacket = 1; } } } }
              
              ## Ignore packets if it's destination host is listed in the "ignoreDstHosts" section of the configuration file.
              unless ($ignorePacket) { foreach my $tmp (@ignoreDstHosts) { if ($dst =~ /$tmp/) { $ignorePacket = 1; } } }
              
              ## Ignore packets if it's source host is listed in the "ignoreSrcHosts" section of the configuration file.
              unless ($ignorePacket) { foreach my $tmp (@ignoreSrcHosts) { if ($src =~ /$tmp/) { $ignorePacket = 1; } } }
              
              
              ## Print this to the log, stdout or whatever
              printmsg("[Packet] src: $src \tdst: $dst\t$protocol port $dpt\tIgnoring: $ignorePacket\tpacket type: $packetType", 0);
              
              
              ## Unless we're ignoring this packet, process it:
              unless ($ignorePacket) {
                  
                  ## If this SRC-DST-PROTO is existing in our database
                  my $hashEntry = $connectionHash{"$src-$dst-$protocol"};
                  if ($hashEntry) {
                      
                      ## If this destination port is a duplicate
                      if ($hashEntry =~ /,$dpt,/) {
                          ## Do nothing - its a duplicate entry - FIXME: Should this update the timestamp for the entry?
                      }
                      
                      ## If this destination port is new
                      else {
                          
                          ## Divide the entry into it's separate parts
                          my @entryPart = split(/:/, $hashEntry);
                          
                          ## Debugging
                          printmsg("entryPart[0]: $entryPart[0]\nentryPart[1]: $entryPart[1]\nentryPart[2]: $entryPart[2]\n\n", 6);
                          
                          ## Update the timestamp
                          $entryPart[0] = time();
                          
                          ## Check to see if entry is in portscan mode:
                          if ($entryPart[1] =~ /1/) {
                              ## If it is, update the entry (add the new port)
                              $connectionHash{"$src-$dst-$protocol"} = $entryPart[0] . ":" . $entryPart[1] . ":" . $entryPart[2] . "$dpt,";
                          }
                          ## Else if it's not in portscan mode:
                          else {
                              ## Append new dpt
                              $entryPart[2] .= "$dpt,";
                              
                              ## See if new port puts us over the threshold
                              my $tmp = $entryPart[2];
                              $tmp =~ s/!(,)//;
                              if (length($tmp) > $conf{"portScanThreshold"}) {
                                  ## If it does, enable the port-scan-in-progress flag
                                  $entryPart[1] = 1;
                                  ## Send syslog messasge
                                  syslog("/dev/log", 4, 3, "Portscan detected! Source: [$src] Destination [$dst] Type [$packetType]");
                              }
                              
                              ## Update the entry
                              $connectionHash{"$src-$dst-$protocol"} = $entryPart[0] . ":" . $entryPart[1] . ":" . $entryPart[2];
                          }
                      }
                  
                  }
                  ## If this SRC-DST-PROTO is new
                  else {
                      ## Create connectionHash entry
                      $connectionHash{"$src-$dst-$protocol"} = time() . ":0:,$dpt,";
                  }
              }
          }
      }
  }
  else {
      ## This tells the parent process to exit if the child forked successfully
      quit("", 0);
  }
  
}



## If MODE is message
elsif ($conf{"mode"} =~ /message/i) {
    
    ## FIXME: Send a message to a running scanAlert process here
    printmsg("ERROR: This feature is not implemented yet!", 0);
    
    
}




## Quit (should never get here)
quit("ERROR: I shouldn't be exiting here!",1);
































######################################################################
## Function:    help ()
##
## Description: For all those newbies ;) 
##              Prints a help message and exits the program.
## 
######################################################################
sub help {
print <<EOM;

$conf{'name'}-$conf{'version'} by $conf{'authorName'} <$conf{'authorEmail'}>

Usage:  $conf{'name'} [options]

 -d,--daemon                        become a daemon
 -c <file>,--config=<file>          specify configuration file
 -v                                 verbosity - use multiple times for greater effect

EOM

## A future command line argument
##   -m <message>,--message=<message>   send a message to a running scanAlert daemon


exit(1);
}










######################################################################
##  Function: initialize ()
##  
##  Does all the script startup jibberish.
##  
######################################################################
sub initialize {

  ## Set STDOUT to flush immediatly after each print  
  $| = 1;

  ## Intercept signals
  $SIG{'QUIT'}  = sub { quit("$$ - EXITING: Received SIG$_[0]", 1); };
  $SIG{'INT'}   = sub { quit("$$ - EXITING: Received SIG$_[0]", 1); };
  $SIG{'KILL'}  = sub { quit("$$ - EXITING: Received SIG$_[0]", 1); };
  $SIG{'TERM'}  = sub { quit("$$ - EXITING: Received SIG$_[0]", 1); };
  
  ## ALARM and HUP signals are not supported in Win32
  unless ($^O =~ /win/i) {
    $SIG{'HUP'}   = sub { quit("$$ - EXITING: Received SIG$_[0]", 1); };
    $SIG{'ALRM'}  = sub { quit("$$ - EXITING: Received SIG$_[0]", 1); };
  }
  
  return(1);
}









######################################################################
##  Function: processCommandLine ()
##  
##  Processes command line storing important data in global var %conf
##  
######################################################################
sub processCommandLine {
  
  
  ############################
  ##  Process command line  ##
  ############################
  
  my $numargv = scalar(@ARGV);
  help() unless ($numargv);
  my $counter = 0;
  for ($counter = 0; $counter < $numargv; $counter++) {

    if ($ARGV[$counter] =~ s/^-d//i) {                   ## Daemon ##
        $conf{'mode'} = "daemon";
    }

    elsif ($ARGV[$counter] =~ s/^--daemon//i) {          ## Daemon ##
        $conf{'mode'} = "daemon";
    }

    elsif ($ARGV[$counter] =~ s/^-m//i) {                ## Message ##
        $conf{'mode'} = "message";
        $counter++;
        $conf{'message'} = $ARGV[$counter];
    }
    
    elsif ($ARGV[$counter] =~ s/^--message=//i) {        ## Message ##
        $conf{'mode'} = "message";
        $conf{'message'} = $';
    }

    elsif ($ARGV[$counter] =~ s/^--config=//i) {         ## Config File ##
        $conf{'configurationFile'} = $';
    }

    elsif ($ARGV[$counter] =~ /^-c$/i) {                 ## Config File ##
        $counter++;
        $conf{'configurationFile'} = $ARGV[$counter];
    }

    elsif ($ARGV[$counter] =~ /^-v$/i) {                 ## Verbosity ##
        $conf{'debug'}++;
    }

    elsif ($ARGV[$counter] =~ /^-h$|^--help$/i) {        ## Help ##
        help();
    }
    
    else {
        printmsg("\"$ARGV[$counter]\" is not a recognised option!", 0);
        help();
    }

  }
  
  ###################################################
  ##  Verify required variables are set correctly  ##
  ###################################################
  my @required = (
                    'configurationFile',
                    'mode',
  );
  foreach (@required) {
      if (!$conf{$_}) {
          quit("ERROR: Value [$_] was not set after parsing command line arguments!", 1);
      }
  }
  
  ## Print help if we're not becoming a daemon.
  if (!$conf{'mode'}) {
      help();
  }
  
  return(1);
}



















###############################################################################################
## FUNCTION:    
##   readConfigurationFile ( string $filename )
## 
## 
## DESCRIPTION: 
##   Reads the file $filename, and stores information from the file into 
##   the global array $conf, and a few other global arrays.  This version
##   of this function is specifically modified to support scanAlert's configuration
##   file.
##
## Example: 
##   readConfigurationFile("/etc/configuration.conf");
##
###############################################################################################
sub readConfigurationFile {
  
  ## Get incoming variables
  my %incoming = ();
  (
      $incoming{'confFile'},
  ) = @_;
  $incoming{'section'} = "";
  
  
  ## Open the configuration file
  open(CONFFILE, $incoming{'confFile'}) or quit("$$ - ERROR While opening the configuration file [$incoming{'confFile'}].  The error was: $!\n",1);
  while (<CONFFILE>) {  ## FOR EACH LINE
      my $line = $_;
      chomp $line;           ## Remove line ending characters
      $line =~ s/\#.+$//;    ## Remove comments
      $line =~ s/\/\/.+$//;  ## Remove comments
      $line =~ s/^\s+//;     ## Remove leading white space
      $line =~ s/\s+$//;     ## Remove ending white space
      
      ## Assign data from the config file to %conf
      if ($line ne "") { 
          $conf{(split(/\s+=>\s+/, $line))[0]} = (split(/\s+=>\s+/, $line))[1];
      }
    
  }
  close(CONFFILE);
  
  
  ## Now we need to convert a few of the entries into arrays
  if ($conf{'ignorePortsTCP'}) { @ignorePortsTCP = split(",", $conf{'ignorePortsTCP'}); }
  if ($conf{'ignorePortsUDP'}) { @ignorePortsUDP = split(",", $conf{'ignorePortsUDP'}); }
  if ($conf{'ignoreDstHosts'}) { 
      $conf{'ignoreDstHosts'} =~ s/\./\\\./;  ## Fixup periods so regular expressions handle them correctly later
      @ignoreDstHosts = split(",", $conf{'ignoreDstHosts'});
  }
  if ($conf{'ignoreSrcHosts'}) { 
      $conf{'ignoreSrcHosts'} =~ s/\./\\\./;  ## Fixup periods so regular expressions handle them correctly later
      @ignoreSrcHosts = split(",", $conf{'ignoreSrcHosts'});
  }
  delete @conf{('ignoreDstHosts','ignoreSrcHosts','ignorePortsTCP','ignorePortsUDP')};
  
  
  ## DEBUGGING:
  printmsg ("Configuration: ", 6); foreach my $key (keys(%conf)) { printmsg ("Key: $key =>\t Value: $conf{$key}", 6) if ($key and $conf{$key}); }
  
  return(1);
}















###############################################################################################
##  Function:    printmsg (string $message, int $level)
##
##  Description: Handles all messages - logging them to a log file, 
##               printing them to the screen or both depending on
##               the $level passed in, $conf{'debug'} and wether
##               $conf{'mode'}.
##
##  Input:       $message          A message to be printed, logged, etc.
##               $level            The debug level of the message. If
##                                 not defined 0 will be assumed.  0 is
##                                 considered a normal message, 1 and 
##                                 higher is considered a debug message.
##  
##  Output:      Prints to STDOUT, to LOGFILE, both, or none depending 
##               on the state of the program.
##  
##  Example:     ("WARNING: We believe in generic error messages... NOT!\n", 1);
###############################################################################################
sub printmsg {
  my %incoming = ();
  (
    $incoming{'message'},
    $incoming{'level'}
  ) = @_;
  $incoming{'level'} = 0 if (!defined($incoming{'level'}));
  
  ## Print to the log file if $conf{'logging'} is true and $conf{'debug'} >= $incoming{'level'}
  if ($conf{'logging'} and ($conf{'debug'} >= $incoming{'level'}) ) {
      print LOGFILE localtime() . " - $incoming{'message'}\n";
  }
  
  ## Print to STDOUT always if debugging is enabled, or if the mode is NOT "running".
  if ( ( ($conf{'debug'} >= 1) or ($conf{'mode'} ne "running") ) and ($conf{'debug'} >= $incoming{'level'}) ) {
      print localtime() . " - $incoming{'message'}\n";
  }
  
  ## Return
  return(1);
}












###############################################################################################
## FUNCTION:    
##   openLogFile ( $filename )
## 
## 
## DESCRIPTION: 
##   Opens the file $filename and attaches it to the filehandle "LOGFILE".  Returns 0 on success
##   and non-zero on failure.  Error codes are listed below, and the error message gets set in
##   global variable $!.
##   
##   
## Example: 
##   openFile ("/var/log/scanAlert.log");
##
###############################################################################################
sub openLogFile {
    ## Get the incoming filename
    my $filename = $_[0];
    
    ## Make sure our file exists, and if the file doesn't exist then create it
    if ( ! -f $filename ) {
        printmsg("NOTICE: The file [$filename] does not exist.  Creating it now with mode [0600].", 0);
        open (LOGFILE, ">>$filename");
        close LOGFILE;
        chmod (0600, $filename);
    }
    
    ## Now open the file and attach it to a filehandle
    open (LOGFILE,">>$filename") or return (1);
    
    ## Put the file into non-buffering mode
    select LOGFILE;
    $| = 1;
    select STDOUT;
    
    ## Tell the rest of the program that we can log now
    $conf{'logging'} = "yes";
    
    ## Return success
    return(0);
}












###############################################################################################
## FUNCTION:    
##   openFIFO ( $filename )
## 
## 
## DESCRIPTION: 
##   Opens the fifo $filename for reading and attaches it to the filehandle "FIFO".  Returns 0 on success
##   and non-zero on failure.  Error codes are listed below, and the error message gets set in
##   global variable $!. 
##   
##
## Example: 
##   openFIFO ("/var/log/fifo.scanAlert");
##
###############################################################################################
sub openFIFO {
    ## Get the incoming filename
    my $filename = $_[0];
    
    ## Make sure our fifo exists, and if the fifo doesn't exist then create it
    if ( ! -p $filename ) {
        quit("ERROR: FIFO does not exist or is not a FIFO at [$conf{'fifo'}]\nDid you forget to create it and configure syslogd to log to it?\n", 1);
    }
    
    ## Open the FIFO socket we're supposed to listen to and attach it to a filehandle
    open (FIFO,"< $filename") or return(1);
    
    ## Return success
    return(0);
}











#########################################################
# FUNCTION: handleHUP()
# 
# Handles HUP signals:
#  - Re-reads the configuration file
#  - Logs a message to the log file
#  - Closes and opens the log file 
#  - Closes and opens the fifo
#
#########################################################
sub handleHUP  {
    
    printmsg("$$ - $conf{'name'} - Received SIGHUP: Reloading config and refreshing all file handles", 0);
    
    ######################################
    ##  Re-read the configuration file  ##
    ######################################
    readConfigurationFile($conf{'configurationFile'});
    
    
    ###################################
    ##  Close and open the log file  ##
    ###################################
    close LOGFILE;
    if (openLogFile($conf{'logFile'})) { quit("ERROR: Opening the file [$conf{'logFile'}] returned the error: $!", 1); }
    
    
    ###############################
    ##  Close and open the fifo  ##
    ###############################
    close FIFO;
    if (openFIFO($conf{'fifo'})) { quit("ERROR: Connecting to fifo [$conf{'fifo'}] returned the error: $!", 1); }
    
    
    ## Go back to the normal routine
    return(1);
}














#########################################################
# FUNCTION: handleALRM ()
# 
# Does periodical maintence for the program: 
#  - Cleans old entries out of the connecitonHash
#  - Writes the connectionHash to a file
#  - Determines if an active port-scan is over?
#
#########################################################
sub handleALRM  {
    
    ###########################################
    ##  Clean out old connectionHash entries ##
    ###########################################
    foreach my $key (keys(%connectionHash)) { 
        
        ## Divide the entry into it's separate parts
        my @entryPart = split(/:/, $connectionHash{$key});
        
        ## If the entry is older than one day, delete it.
        if ($entryPart[0] < (time() - $conf{'hostExpiration'})) {
            
            ## If the entry was in port-scan-mode, first log it's final data
            if ($entryPart[1]) {
              (my $src, my $dst, my $protocol) = split("-", $key);
              syslog("/dev/log", 4, 3, "$protocol Portscan complete:  Source [$src] -> Destination [$dst]  Ports scanned [$entryPart[2]]");
            }
            delete $connectionHash{$key} 
        }
    }
    
    
    ##################################################
    ##  Write connectionHash to the stateData file  ##
    ##################################################
    
    if ( ! -f "$conf{'stateData'}" ) { 
        printmsg("NOTICE: The file [$conf{'stateData'}] does not exist.  Creating it now with mode [0600].", 0);
        open (FILE, ">$conf{'stateData'}");
        close FILE;
        chmod (0600, $conf{'stateData'});
    }
    open (FILE, ">$conf{'stateData'}");
    foreach my $key (keys(%connectionHash)) { 
        print FILE "$key\t=> $connectionHash{$key}\n";
    }
    close FILE;
    
    
    
    ## Go back to the normal routine
    return(1);
}














###############################################################################################
## FUNCTION:    
##   getPacketInfo (string $iptables log entry)
## 
## 
## DESCRIPTION: 
##   Parses a standard IPTables syslog entry and returns information about the packet.
##   It returns an array of data like this:
##     (  
##        Source IP, 
##        Destination IP, 
##        Protocol,
##        Destination Port,
##        Packet Type
##     )
##
## Example: 
##   getPacketInfo ("Feb 20 13:06:49 system kernel: IN=eth0 OUT= MAC=00:00:00:00:00:00:00:00:11:22:33:44:55:66 SRC=192.168.57.77 DST=192.168.91.194 LEN=44 TOS=0x00 PREC=0x00 TTL=119 ID=35118 DF PROTO=TCP SPT=4461 DPT=80 WINDOW=8192 RES=0x00 SYN URGP=0");
##
###############################################################################################
sub getPacketInfo {
  
  
  ## Get incoming variables
  my $packet = $_[0];
  
  
  ## Define packet types - NOTE the order of TCP flags should be: ACK URG PSH FIN RST SYN
  my %scanType = (
      ",SYN"           => "Normal/SYN",
      ",FIN"           => "Stelth FIN",
      ",URG,PSH,FIN"   => "Xmas",
      ""               => "Null",
      ",ACK,PSH,FIN"   => "Expired Connection",
  );
  
  
  ## Declare some variables
  my $src = "";          ## Source host
  my $dst = "";          ## Destination host
  my $dpt = "";          ## Destination port
  my $protocol = "";     ## Protocol
  my $tcpFlags = "";     ## TCP Flags
  my $packetType = "";   ## Packet type (derived from tcp flags)
  
  
  ## Find protocol, DST port, source IP, dest IP
  $packet =~ s/SRC=(\d+)\.(\d+)\.(\d+)\.(\d+)//;
  $src = $&; $src =~ s/SRC=//;
  
  $packet =~ s/DST=(\d+)\.(\d+)\.(\d+)\.(\d+)//;
  $dst = $&; $dst =~ s/DST=//;
  
  $packet =~ s/DPT=\d+//;
  $dpt = $&; $dpt =~ s/DPT=//;
  
  $packet =~ s/PROTO=\w+//;
  $protocol = $&; $protocol =~ s/PROTO=//;
  
  
  ## Now find which TCP flags were used (ACK, URG, PSH, FIN, RST, SYN)
  if ($protocol eq "TCP") {
      
      $tcpFlags .= ",ACK" if ($packet =~ / ACK /);
      $tcpFlags .= ",URG" if ($packet =~ / URG /);
      $tcpFlags .= ",PSH" if ($packet =~ / PSH /);
      $tcpFlags .= ",FIN" if ($packet =~ / FIN /);
      $tcpFlags .= ",RST" if ($packet =~ / RST /);
      $tcpFlags .= ",SYN" if ($packet =~ / SYN /);
      
      $packetType = $scanType{$tcpFlags};
      if (!$packetType) {
          $packetType = "No known configuration matches TCP flags [$tcpFlags]";
      }
  
  } else {
      $packetType = "UDP Packet";
  }
  
  
  ## Return information
  return($src, $dst, $protocol, $dpt, $packetType);
}















###############################################################################################
## FUNCTION:    
##   syslog (string $socketName, int $priority, int $facility, string $message)
## 
## 
## DESCRIPTION: 
##   Connects to the socket $socketName, and sends it a syslog formatted message.  
##   i.e. it syslog's $message with the priority and facility specified.
##
## Example: 
##   syslog("/dev/log", 6, 1, "Test syslog message");
##
##
## Priorities (on a Linux system)
##   LOG_EMERG       0       system is unusable
##   LOG_ALERT       1       action must be taken immediately
##   LOG_CRIT        2       critical conditions
##   LOG_ERR         3       error conditions
##   LOG_WARNING     4       warning conditions
##   LOG_NOTICE      5       normal but significant condition
##   LOG_INFO        6       informational
##   LOG_DEBUG       7       debug-level messages
##
##
## Facilities (on a Linux system)
##   LOG_KERN        0       kernel messages
##   LOG_USER        1       random user-level messages
##   LOG_MAIL        2       mail system
##   LOG_DAEMON      3       system daemons
##   LOG_AUTH        4       security/authorization messages
##   LOG_SYSLOG      5       messages generated internally by syslogd
##   LOG_LPR         6       line printer subsystem
##   LOG_NEWS        7       network news subsystem
##   LOG_UUCP        8       UUCP subsystem
##   LOG_CRON        9       clock daemon
##   LOG_AUTHPRIV    10      security/authorization messages (private)
##   LOG_FTP         11      ftp daemon
##
###############################################################################################
sub syslog {
  
  ## Get incoming variables
  my %incoming = ();
  (
    $incoming{'socketName'},
    $incoming{'priority'},
    $incoming{'facility'},
    $incoming{'message'},
  ) = @_;
  
  ## Set defaults if some variables are not defined
  $incoming{'socketName'}  = "/dev/log"         if (!defined($incoming{'socketName'}));
  $incoming{'priority'}    = 6                  if (!defined($incoming{'priority'}));
  $incoming{'facility'}    = 1                  if (!defined($incoming{'facility'}));
  $incoming{'message'}     = "Default Message"  if (!defined($incoming{'message'}));
  my $return = 1;

  ## Make sure values are sane (black magic)
  $incoming{'priority'} = $incoming{'priority'} & 0x07;
  $incoming{'facility'} = $incoming{'facility'} & 0x3f8;
  
  
  ## Debuging
  printmsg("syslog(): Syslogging the message [$incoming{'message'}]", 1);
  
  ## Open a UNIX socket in dgram mode with udp protocol. 
  socket(SOCKET, PF_UNIX, SOCK_DGRAM, 0) || ($return = 0);
  if ($return == 0) { printmsg("$$ - syslog() - Error: $!", 0); }
  
  ## Connect our socket to SOCKET
  connect(SOCKET, sockaddr_un ($incoming{'socketName'})) || ($return = 0);
  if ($return == 0) { printmsg("$$ - syslog() - Error: $!", 0); }
  
  
  
  ## Generate syslog value for the priority and facility (black magic)
  $incoming{'value'} = ($incoming{'priority'} + (8 * $incoming{'facility'}));

  ## Sending the message to the syslog socket
  print SOCKET "<$incoming{'value'}>$incoming{'message'}" || ($return = 0);
  if ($return == 0) { printmsg("$$ - syslog() - Error: $!", 0); }
  
  
  ## Close the socket
  close SOCKET;
  
  
  ## Return success
  return($return);
}











######################################################################
##  Function:    quit (string $message, int $errorLevel)
##  
##  Description: Exits the program, optionally printing $message.  It 
##               returns an exit error level of $errorLevel to the 
##               system  (0 means no errors, and is assumed if empty.)
##
##  Example:     quit("Exiting program normally", 0);
######################################################################
sub quit {
  my %incoming = ();
  (
    $incoming{'message'},
    $incoming{'errorLevel'}
  ) = @_;
  $incoming{'errorLevel'} = 0 if (!defined($incoming{'errorLevel'}));
  
  ## Write stateData to a file
  handleALRM() if ($conf{"mode"} =~ /running/i);
  
  ## Print exit message
  if ($incoming{'message'}) { 
    printmsg($incoming{'message'}, 0);
    if ($incoming{'errorLevel'} >= 0) {
        syslog("/dev/log", 6, 1, $incoming{'message'});
    }
  }
  
  ## Exit
  exit($incoming{'errorLevel'});
}






